"""
QMS scenario generator v2.0. Copyright Unigraf 2024

The script generates playback content (scenarios and playlists) for QMS transition teste HF2-73, HFR2-73,
and QMS tolerance tests HF2-75, HFR2-75.

Video frames are shared between scenarios. They are stored under 'shared-video' folder.
This folder has a sub-folder 'ppms' for ppm files generated with qms_vrr_tpg.exe.
Test-specific parts are stored under 'HF[R]2-73' and 'HF[R]2-75' folders.
Packets and scenarios depend on the selected VIC. They are stored under VIC-specific folders.
Each VIC-specific folder also contains two playlists:
- groups_list.txt
- transitions_list.txt

The first one plays TFR transitions grouped by the lower frame rate, e.g. 1-2-1-3-1-4.. This type of playlist
is recommended for faster validation.
The second playlist plays all TFR combinations separately, e.g. 1-2, 1-3, 2-3, etc. This is recommended if
the fast testing shows instability in video or audio. To narrow down the faulty TFR combinations, the second playlist
and its scenarios should be used.

root - the current directory
├─ vic-63-RGB444\
│  ├─ shared-video\
│  │  ├─ ppms\
│  │  │  ├─ qms_test_00*.ppm
│  │  ├─ packet*.bin
│  │  ├─ qms_test_00*.jpg
│  ├─ HF[R]2-73\
│  │  ├─ groups_list.txt
│  │  ├─ transitions_list.txt
│  │  ├─ groups\
│  │  │  ├─ scenario-all.txt
│  │  │  ├─ scenario-all-0.txt
│  │  │  ├─ scenario-group3_5.3s.txt
│  │  │  ├─ ...
│  │  ├─ step-1-2\
│  │  │  ├─ scenario.txt
│  │  │  ├─ scenario-0.txt
│  │  │  ├─ packet*.bin
│  │  │  ├─ qms_test_00*.jpg
│  │  ├─ step-1-3\
│  │  │  ├─ scenario.txt
│  │  │  ├─ scenario-0.txt
│  │  │  ├─ packet*.bin
│  │  │  ├─ qms_test_00*.jpg
│  │  ├─ step-1-4...
│  │  │  ├─ ...
│  ├─ HF[R]2-75
│  │  ├─ groups_list.txt
│  │  ├─ transitions_list.txt
│  │  ├─ packet*.bin
│  │  ├─ groups\
│  │  │  ├─ scenario-all.txt
│  │  │  ├─ scenario-all-0.txt
│  │  │  ├─ scenario-group3_5.3s.txt
│  │  │  ├─ ...
│  │  ├─ step-1-2\
│  │  │  ├─ scenario.txt
│  │  │  ├─ scenario-0.txt
│  │  │  ├─ packet*.bin
│  │  │  ├─ qms_test_00*.jpg

"""
import os
import subprocess
import shutil
import glob
from pathlib import Path
import cv2
import time

version = 'v2.0'

# The values for UCD-422 with 2GB DDR on board memory
max_2K_frames = 72
max_4K_frames = 36
max_8K_frames = 9

# to generate shorter scenarios, set this True
shorter = False

# to generate all steps, set this True
steps = True

sub_folder_transition = 'HF[R]2-73'
sub_folder_tolerance = 'HF[R]2-75'
sub_folder_common = 'shared-video'
audio_file = '44100Hz_16bit_PCM_2channels_30sec_sine.wav'

# For AVI packet
# NOTE: a scenario file matching this VIC should exist e.g. scenario78.txt
timing_id = 78  # VIC
V_Total = None
color_space_idx = 0  # RGB, YUV422, YUV444, YUV420

# For playing order
COLOUR_FORMAT = '256'  # RGB 8

QMS_RATES = [0, 24 / 1.001, 24.0, 25.0, 30 / 1.001, 30.0, 48 / 1.001, 48.0, 50.0, 60 / 1.001, 60.0, 100.0, 120 / 1.001,
             120.0]
BASE_RATE = None  # Will be replaced with user input; should match the scenario settings

"""
BAR_WIDTHS   0,2.083,6.25,3.125   // No bar on top 1/4th of screen, only 2nd through 4th values are used
MOTION_AMOUNTS 0,12.5,-12.5,33.333   // No bar on top 1/4th of screen, only 2nd through 4th values are used
"""
amounts_float = [0, 100 / 36 / 2.083 * 100, -100 / 12 / 6.25 * 100, 100 / 18 / 3.125 * 100]
amounts = ','.join([f'{x:.2f}' for x in amounts_float])


def load_packet_bin(file_name):
    try:
        b = open(file_name, 'rb')
    except IOError:
        print('Failed to read ' + file_name)
        return None

    res = b.read(36)
    return bytearray(res)


def save_packet_bin(info, root, packet_index, buffer):
    path = root + os.sep + f'packet{packet_index}.bin'
    try:
        b = open(path, 'wb')
    except IOError:
        print('Failed to write ' + path)
        return False

    log.write(f'\nWriting packet: {info} to "{path}"')
    return b.write(bytearray(buffer))


def checksum_hdmi(b):
    s = 0
    b[3] = 0
    sz = len(b)
    for i in range(0, sz):
        s += b[i]
    b[3] = (~s + 1) & 0xFF
    # let's check
    s = 0
    for i in range(0, sz):
        s += b[i]

    s &= 0xFF
    if s != 0:
        print("HDMI Checksum failure: 0x" + str(s))
    return s == 0


def copy_avi_packet(file_name, packet_index, dest):
    packet = load_packet_bin(file_name)
    if not packet or not len(packet):
        return False
    colors = [0, 0x40, 0x20, 0x60]  # RGB, YUV422, YUV444, YUV420
    packet[4] |= colors[color_space_idx]
    packet[7] = timing_id  # VIC
    if not checksum_hdmi(packet):
        return False
    return save_packet_bin(f'AVI (VIC={timing_id})', dest, packet_index, packet)


def copy_qms_emp_packet(file_name, tfr, M_Const, packet_index, dest):
    packet = load_packet_bin(file_name)
    if not packet or not len(packet):
        return False

    # Refer to HDMI 2.1a specs, Table 10-37

    # Update tfr and M_CONST
    if M_Const:
        packet[0xA] |= 2
    else:
        packet[0xA] &= (~2 & 0xFF)

    # Disable QMS if tfr is zero
    if not tfr:
        packet[0xA] &= (~4 & 0xFF)

    # Set base frame rate and Next_TFR
    fr = int(BASE_RATE + 0.5)
    packet[0xC] = ((tfr << 3) | (fr >> 8)) & 0xFF
    packet[0xD] = fr & 0xFF

    return save_packet_bin(f'QMS EMP M_CONST={M_Const} Next_TFR={tfr}', dest, packet_index, packet)


def tfr_to_string(tfr, adjust_m):
    rate0 = QMS_RATES[tfr]
    if rate0 > BASE_RATE:  # Should never happen
        print('Warning: QMS rate ', rate0, ' exceeds base frame rate ', BASE_RATE)
        rate0 = BASE_RATE

    v0 = int(rate0 * adjust_m)
    m_vrr = int(int(V_Total) * (BASE_RATE - rate0) / rate0) if v0 else 0
    return 'm' + str(m_vrr) if m_vrr > 0 else ''


def playing_order(qms, adjust_m=1):
    #  For each TFR part, M_CONST: [0][1][1]..[1][1]
    #  Corresponding frame rate values are shifted. Previous frame rate should be set with 'v' for the M_CONST[0] frame.
    #  For the first TFR part, the previous should be taken from the last part
    prevM = tfr_to_string(start, adjust_m) if qms and start else ''
    nextM = tfr_to_string(end, adjust_m) if qms and end else ''

    # For example
    # const = 1   VRR = 48
    # const = 0   VRR = 120
    # const = 1   VRR = 120
    # const = 0   VRR = 48
    if qms:
        porder = f'0-{transition_frame - 1}:1:0,1:' + COLOUR_FORMAT + prevM + "; "
        porder += f'{transition_frame}:1:0,2:' + COLOUR_FORMAT + prevM + "; "
        porder += f'{transition_frame + 1}-{last_frame - 1}:1:0,3:' + COLOUR_FORMAT + nextM + "; "
        porder += f'{last_frame}:1:0,4:' + COLOUR_FORMAT + nextM + ";"
    else:
        porder = f'0-{last_frame}:1:0:' + COLOUR_FORMAT + ";"

    return porder


def playing_order_all(qms, adjust_m=1):
    # For example
    # const = 1   VRR = 48
    # const = 0   VRR = 120
    # const = 1   VRR = 120
    # const = 0   VRR = 48
    porder = ''
    last_frame = num_frames - 1
    if qms:
        for start in range(minTFR, maxTFR + 1):
            for end in range(start + 1, maxTFR + 1):
                prevM = tfr_to_string(start, adjust_m) if start else ''
                nextM = tfr_to_string(end, adjust_m) if end else ''
                porder += f'0-{last_frame - 1}:1:0,{(start - minTFR) * 2 + 1}:' + COLOUR_FORMAT + prevM + "; "
                porder += f'{last_frame}:1:0,{(end - minTFR) * 2 + 2}:' + COLOUR_FORMAT + nextM + "; "
                porder += f'0-{last_frame - 1}:1:0,{(end - minTFR) * 2 + 1}:' + COLOUR_FORMAT + nextM + "; "
                porder += f'{last_frame}:1:0,{(start - minTFR) * 2 + 2}:' + COLOUR_FORMAT + prevM + ";"
    else:
        porder = f'0-{last_frame}:1:0:' + COLOUR_FORMAT + ";"

    return porder


def group_time(start):
    total = 0.0
    brr = QMS_RATES[start]
    for end in range(start + 1, maxTFR + 1):
        rate = QMS_RATES[end]
        total += num_frames / brr
        total += num_frames / rate

    return total


def playing_order_group(start, adjust_m=1):
    # For example
    # const = 1   VRR = 48
    # const = 0   VRR = 120
    # const = 1   VRR = 120
    # const = 0   VRR = 48
    porder = ''
    last_frame = num_frames - 1
    for end in range(start + 1, maxTFR + 1):
        prevM = tfr_to_string(start, adjust_m) if start else ''
        nextM = tfr_to_string(end, adjust_m) if end else ''
        porder += f'0-{last_frame - 1}:1:0,{(start - minTFR) * 2 + 1}:' + COLOUR_FORMAT + prevM + "; "
        porder += f'{last_frame}:1:0,{(end - minTFR) * 2 + 2}:' + COLOUR_FORMAT + nextM + "; "
        porder += f'0-{last_frame - 1}:1:0,{(end - minTFR) * 2 + 1}:' + COLOUR_FORMAT + nextM + "; "
        porder += f'{last_frame}:1:0,{(start - minTFR) * 2 + 2}:' + COLOUR_FORMAT + prevM + ";"

    return porder


def copy_scenario(source, content_folder, audio_folder, porder, destination, adjust_frame_rate):
    try:
        src = open(source, "r")
    except IOError as e:
        print(f'Failed to open {source} error {str(e)}')
        return None

    lines = src.readlines()
    src.close()

    try:
        dst = open(destination, "w")
    except IOError as e:
        print(f'Failed to open {destination} error {str(e)}')
        return None

    for line in lines:
        if line.find('porder=') == 0:
            line = 'porder=' + porder + '\n'
        elif line.find('audio=') == 0:
            line = 'audio=../../../' + audio_folder + '/' + audio_file + '\n'
        elif line.find('video=') == 0:
            line = 'video=../../../' + content_folder + '/qms_test_000000.jpg\n'
        elif line.find('packets=') == 0:
            line = 'packets=../../../' + content_folder + '/packet0.bin\n'
        elif adjust_frame_rate != 1 and line.find('timing.frate') == 0:
            parts = line.split('=')
            if len(parts) == 2:
                current_rate = int(parts[1].strip())
                new_frame_rate = int(adjust_frame_rate * current_rate)
                line = f'timing.frate={new_frame_rate}\n'
                rate_message = f'Frame rate adjusted from {current_rate} to {new_frame_rate}'
                # print(rate_message)
                log.write('\n' + rate_message + '\n')

        dst.write(line)

    dst.close()
    return destination


def scenario_info(name):
    if name.find('scenario') != 0:
        return None, None, None, None, None, None
    vic = name[len('scenario'):-4]
    dash = vic.find('-')
    if dash >= 0:
        vic = vic[:dash]
    with open(name, 'r') as f:
        lines = f.readlines()
        # default values should be replaced with data from scenario file
        format = 'RGB'
        vtotal = 1125
        rate = 120
        h_active = 1920
        v_active = 1080
        rate = 60

        for line in lines:
            if line.find('colorspace') == 0:
                format = line[len('colorspace='):-1]
            elif line.find('timing.hactive') == 0:
                h_active = line[len('timing.hactive='):-1]
            elif line.find('timing.vactive') == 0:
                v_active = line[len('timing.vactive='):-1]
            elif line.find('timing.vtotal') == 0:
                v_total = line[len('timing.vtotal='):-1]
            elif line.find('timing.frate') == 0:
                rate = line[len('timing.frate='):-1]

        return vic, format, v_total, h_active, v_active, rate


def label(file, upper, lower, transition, destination):
    img = cv2.imread(file, cv2.IMREAD_COLOR) * 16384
    font = cv2.FONT_HERSHEY_COMPLEX
    # get boundary of this text
    text_size = cv2.getTextSize(upper, font, 2, 3)[0]
    text_x = (img.shape[1] - text_size[0]) / 2
    cv2.putText(img, upper, (int(text_x), 500), font, 2,
                (0 if transition else 32000, 32000, 32000), 3)

    if lower:
        text_size = cv2.getTextSize(lower, font, 2, 3)[0]
        text_x = (img.shape[1] - text_size[0]) / 2
        cv2.putText(img, lower, (int(text_x), 600), font, 2,
                    (32000, 32000, 32000), 3)
    elif '=' in upper:
        # get the frame number from the upper label
        parts = upper.split('=')
        bits = 8
        back = (15, 15, 15)
        if len(parts) == 2:
            number = int(parts[1])
            side = 60
            gap = 10
            half = bits / 2 * (side + gap) + gap / 2
            left = int(960 - half)
            right = int(960 + half)
            y2 = 600
            y1 = 600 - side - gap * 2
            cv2.rectangle(img, (left, y1), (right, y2), back, -1)
            y1 += gap
            y2 -= gap
            x1 = int(left + gap)
            for b in range(bits):
                x2 = x1 + side
                color = back if ((128 >> b) & number) == 0 else (200, 200, 200)
                cv2.rectangle(img, (x1, y1), (x2, y2), color, -1)
                x1 += side + gap

    p = Path(file)
    jpg = destination + os.sep + p.stem + '.jpg'
    cv2.imwrite(jpg, img)


def add_labels(folder, show_fps, transition_frame, start, end, destination, mark):
    files = glob.glob('qms_test_*.ppm', root_dir=folder)
    count = 0
    for file in files:
        count += 1
        if show_fps:
            tfr = start if count < transition_frame else end
            transition = count < 3 if count <= transition_frame else count < transition_frame + 3
            label(str(folder) + os.sep + file, f'{mark} Frame={count:03}', f'{QMS_RATES[tfr]:.2f}fps TFR={tfr}',
                  transition, destination)
        else:
            label(str(folder) + os.sep + file, f'{mark} Frame={count:03}', '', count < 3, destination)


def prepare_folder(folder_name, erase):
    _folder = Path(folder_name)
    if erase:
        if _folder.is_dir():
            try:
                shutil.rmtree(_folder)
            except OSError as er:
                print("Error: %s - %s." % (er.filename, er.strerror))
                return False

    if not _folder.is_dir():
        try:
            _folder.mkdir()
        except OSError as e:
            print(f'Failed to create folder {str(_folder)} reason {str(e)}')
            return False

    return True


def copy_all_emp(destination):
    idx = 1   # the 0 is AVI
    for tfr in range(minTFR, maxTFR + 1):
        copy_qms_emp_packet('qms.bin', tfr, True, idx, destination)
        idx += 1
        copy_qms_emp_packet('qms.bin', tfr, False, idx, destination)
        idx += 1


def write_playlist(filename, playlist):
    with open(filename, 'w') as f:
        f.write('[Device]\nserialnumber = *\nconnectortype = HDMI Out\nreset = 0\nstrict = 0\n\n')
        for key in playlist:
            f.write('[Scenario]\n')
            f.write(f'caption = {playlist[key]}\n')
            f.write('period = 300000\n')
            f.write('prompt = Record video for analysis with high-speed camera. Check audio.<br>Click OK to continue.\n')
            f.write(f'path = {key}\n\n')


def select_scenario():
    files = glob.glob('scenario*.txt')
    count = 0
    for file in files:
        vic, fmt, V_Total, width, height, rate = scenario_info(file)
        if vic:
            count += 1
            print(f'{count}. VIC={vic:3} {width:4}x{height:4} {fmt} {float(rate) / 1000:6}')

    if not count:
        return None

    while True:
        entry = int(input(f'Select the option [1..{len(files)}] '))
        if 0 < entry <= len(files):
            break

    return files[entry - 1]


if __name__ == '__main__':
    caption = f'QMS scenario generator {version}. Copyright Unigraf 2024\n'
    print(caption)

    scenario = select_scenario()
    if not scenario:
        print('Nothing to do. Are scenario files missing?')
        exit(-999)

    max_tfr = 13
    max_frames = 0
    vic, fmt, V_Total, width, height, rate = scenario_info(scenario)
    BASE_RATE = float(rate) / 1000
    while BASE_RATE < QMS_RATES[max_tfr]:
        max_tfr -= 1
    v_total = int(V_Total)
    max_frames = max_2K_frames if v_total < 2000 else max_4K_frames if v_total < 4000 else max_8K_frames
    print(f'BRR {BASE_RATE} -> maximum TFR: {max_tfr} ({QMS_RATES[max_tfr]}) number of frames: {max_frames}')

    while True:
        minTFR = int(input(f'Minimum TFR [1..{max_tfr - 1}] '))
        if 0 < minTFR <= max_tfr - 1:
            break

    while True:
        min1 = minTFR + 1
        maxTFR = int(input(f'Maximum TFR [{min1}..{max_tfr}] '))
        if min1 <= maxTFR <= max_tfr:
            break

    num_frames = max_frames / 2 if shorter else max_frames

    steps = (input("Generate separate TFR pairs? (y/n) ") + '-').lower().strip()[0] == 'y'
    print('Grouped scenarios for TFR transitions will be generated.')
    if steps:
        print('Separate scenarios for all TFR combinations will be generated.')

    #    LINK_RATE = int(input('FRL Link (1-6), or 0 for TMDS '))

    vic_dir = Path('vic-' + vic + '-' + fmt)
    if not prepare_folder(vic_dir, True):
        exit(-1)

    vic_base = str(vic_dir) + os.sep
    vic_dir_len = len(str(vic_base))   # we'll use it later
    out_folder_common = vic_base + sub_folder_common
    out_folder_transition = vic_base + sub_folder_transition
    out_folder_tolerance = vic_base + sub_folder_tolerance

    msg = f'Scenarios for TFR=[{minTFR}..{maxTFR}]: number of frames {num_frames}'
    msg += f'\nBase frame rate {BASE_RATE}'
    msg += f'\nCommon destination folder {out_folder_common}'
    msg += f'\nHF2-73 destination folder {out_folder_transition}'
    msg += f'\nHF2-75 destination folder {out_folder_tolerance}'
    print(msg)
    log = open('generator.log', 'w')
    log.write(f'{caption}TFR from {minTFR} to {maxTFR} '
              f'{msg}\n')

    folders = [out_folder_common, out_folder_common + os.sep + 'ppms',
               out_folder_transition, out_folder_tolerance]
    for out_folder in folders:
        if not prepare_folder(out_folder, False):
            exit(-1)

    print('Working... Please wait...')

    started = time.time()

    try:
        dest = shutil.copy(audio_file, out_folder_common)
    except IOError as e:
        print(f'Failed to copy {audio_file} reason {str(e)}')
        exit(-11)

    print(f'Copied {dest}')
    log.write(f'Audio {dest}\n')

    log.write('\nMOTION AMOUNTS=' + amounts + '\n')
    half = num_frames // 2
    duration1 = half / QMS_RATES[1]
    duration2 = half / QMS_RATES[1]
    frame = f'{duration1:.03}s@TFR1,{duration2:.03}s@TFR1'
    ppms_folder = str(out_folder_common) + os.sep + 'ppms'
    print(f'Writing PPM images to "{ppms_folder}" ...')
    completed = subprocess.run(["qms_vrr_tpg.exe", '-F', 'config.cfg',
                                '-O', 'OUT_DIR ' + ppms_folder,
                                '-O', 'FRAME_PATTERN ' + frame, '-O', 'MOTION_AMOUNTS ' + amounts],
                               stdout=subprocess.PIPE)

    output = completed.stdout.decode('utf-8')
    log.writelines([x for x in output.split('\n') if x])
    if completed.returncode:
        print('Failure in common : ', completed)
        exit(completed.returncode)

    # copies ppm images to shared-video, adding labels and saving as jpg
    add_labels(ppms_folder, False, 0, 0, 0, out_folder_common, 'QMS-VRR')

    # The same packets are used for both HF2-73 and HF2-75 tests
    vic, fmt, V_Total, width, height, rate = scenario_info(scenario)
    BASE_RATE = float(rate) / 1000
    timing_id = int(vic)
    formats = {'RGB444': 0, 'YUV422': 1, 'YUV444': 2, 'YUV420': 3}
    playing_formats = {'RGB444': '256', 'YUV422': '770', 'YUV444': '257', 'YUV420': '259'}
    if fmt in playing_formats:
        COLOUR_FORMAT = playing_formats[fmt]
    else:
        print('Not expected colour format', fmt, 'in', scenario)
        exit(-12)

    if fmt in formats:
        color_space_idx = formats[fmt]
    else:
        print('Not expected colour format', fmt, 'in', scenario)
        exit(-13)

    if not copy_avi_packet("avi.bin", 0, out_folder_common):
        exit(-4)

    copy_all_emp(out_folder_common)

    # The transition HD2-73 and tolerance HF2-75 tests are different in adjustment of Mvrr and base frame rate.
    tests = [{'title': 'HF[R]2-73 Transition', 'adjust_m': 1, 'adjust_rate': 1,
              'folder': out_folder_transition, 'prefix': 'transition-', 'extra': '[2-73]'},
             {'title': 'HF2-75 Tolerance step 3', 'adjust_m': 1.001, 'adjust_rate': 0.995,
              'folder': out_folder_tolerance, 'prefix': 'step3-', 'extra': '[2-75]'},
             {'title': 'HF2-75 Tolerance step 4', 'adjust_m': 0.999, 'adjust_rate': 1.005,
              'folder': out_folder_tolerance, 'prefix': 'step4-', 'extra': '[2-75]'}]

    for test in tests:
        test_folder = test['folder']
        adjust_rate = test['adjust_rate']
        adjust_m = test['adjust_m']
        prefix = test['prefix']
        title = test['title']
        extra = test['extra']

        group_folder = test_folder + os.sep + 'groups'
        msg = f'=== {title} group scenario folder "{group_folder}" adjust M={adjust_m} R={adjust_rate}'

        print(msg)
        log.write(f'\n\n"{msg}"\n')
        if not prepare_folder(group_folder, False):
            exit(-2)

        p = playing_order_all(True, adjust_m)

        base = group_folder + os.sep + prefix
        name = copy_scenario(scenario, out_folder_common, out_folder_common, p, base + 'scenario-all.txt', adjust_rate)
        if not name:
            exit(-9)
        log.write(f'\nWritten QMS scenario to "{name}" with playing order="{p}"')
        p = playing_order_all(False)
        name = copy_scenario(scenario, out_folder_common, out_folder_common, p, base + 'scenario-all-0.txt', adjust_rate)
        if not name:
            exit(-9)
        log.write(f'\nWritten none-QMS scenario to "{name}" with playing order="{p}"')

        playlist = dict()
        count = 1
        number = maxTFR - minTFR
        for start in range(minTFR, maxTFR):
            p = playing_order_group(start)
            total = group_time(start)
            filename = f'{prefix}scenario-group{start}_{total:.1f}s.txt'
            name = copy_scenario(scenario, out_folder_common, out_folder_common, p, group_folder + os.sep + filename, adjust_rate)
            if not name:
                exit(-9)
            log.write(f'\nWritten QMS scenario to "{name}" with playing order="{p}"')
            # Exclude the leading vic folder from the scenario path
            playlist[name[vic_dir_len:]] = f'{title} Group scenario {start}-{maxTFR} {total:.1f}s [{count}/{number}]'
            count += 1

        write_playlist(str(vic_dir) + os.sep + prefix + 'list_groups.txt', playlist)

        playlist.clear()

        if steps:
            count = 1
            number = (maxTFR - minTFR) * (maxTFR - minTFR + 1)
            for start in range(minTFR, maxTFR):
                for end in range(start + 1, maxTFR + 1):
                    step_name = 'step-' + str(start) + '-' + str(end)
                    step_folder = test_folder + os.sep + step_name
                    print(f'Current scenario folder "{step_folder}"')
                    log.write(f'\nScenario folder: "{step_folder}"\n')
                    if not prepare_folder(step_folder, False):
                        exit(-2)

                    transition_frame = half
                    last_frame = num_frames - 1

                    add_labels(ppms_folder,True, transition_frame, start, end, step_folder, extra)

                    # Copy AVI and QMS packets
                    if not copy_avi_packet("avi.bin", 0, step_folder):
                        exit(-4)

                    # AVI packet has index 0
                    # These types of QMS EMP packets are needed:
                    # - M_CONST = 1, Next_TFR = start
                    # - M_CONST = 0, Next_TFR = end (at the transition frame)
                    # - M_CONST = 1, Next_TFR = end
                    # - M_CONST = 0, Next_TFR = start (at the last frame of scenario)

                    if not copy_qms_emp_packet("qms.bin", start, True, 1, step_folder):
                        exit(-5)
                    if not copy_qms_emp_packet("qms.bin", end, False, 2, step_folder):
                        exit(-6)
                    if not copy_qms_emp_packet("qms.bin", end, True, 3, step_folder):
                        exit(-7)
                    if not copy_qms_emp_packet("qms.bin", start, False, 4, step_folder):
                        exit(-8)

                    p = playing_order(True, adjust_m)
                    base = step_folder + os.sep + prefix
                    name = copy_scenario(scenario, step_folder, out_folder_common, p, base + 'scenario.txt', adjust_rate)
                    if not name:
                        exit(-9)
                    log.write(f'\nWritten QMS scenario to "{name}" with playing order="{p}"')
                    playlist[name[vic_dir_len:]] = f'{title} TFR {start}-{end} [{count} / {number}]'
                    count += 1
                    p = playing_order(False)
                    name = copy_scenario(scenario, step_folder, out_folder_common, p, base + 'scenario-0.txt', adjust_rate)
                    if not name:
                        exit(-9)
                    log.write(f'\nWritten none-QMS scenario to "{name}" with playing order="{p}"')

            write_playlist(str(vic_dir) + os.sep + prefix + 'list_transitions.txt', playlist)

    elapsed = f'Done in {time.time() - started:.1f} seconds.'
    print(elapsed)
    log.write('\n\n' + elapsed + '\n')

    log.close()

